@@ -6,15 +6,12 @@ from pubchempy import get_compounds
 from math import pi, exp, ceil
 from matplotlib import pyplot 
 from itertools import chain
+from pprint import pprint
 from sigfig import round
 from glob import glob
 import datetime
 import pandas
-import sigfig
-from pprint import pprint
-import json
-import os
-import re
+import json, os, re
 
 def sigfigs_conversion(num, sigfigs_in = 2):
     return round(num, sigfigs=sigfigs_in, notation = 'sci')
@@ -38,6 +35,16 @@ def time_determination(time):
             unit = 'minutes'
     return sigfigs_conversion(time, 3), unit
 
+def isnumber(num):
+    try:
+        float(num)
+        return True
+    except:
+        try:
+            int(num)
+            return True
+        except:
+            return False
 
 class ROSSPkg():
     def __init__(self, operating_system = 'windows', verbose = False, jupyter = False):      
@@ -58,7 +65,7 @@ class ROSSPkg():
         self.phreeqc_mod = phreeqc_mod
         self.parameters['os'] =  operating_system
         self.parameters['root_path'] = os.path.join(os.path.dirname(__file__))
-        self.databases = [re.search('(\w+)(?=.json)', database).group() for database in glob('./databases/*.json')]
+        self.databases = [re.search('(?<=databases\\\\)(.+)(?=\.json)', database).group() for database in glob(os.path.join(self.parameters['root_path'], 'databases', '*.json'))]
         
     def define_general(self, database_selection, simulation = 'scaling', domain = 'single', domain_phase = None, quantity_of_modules = 1, simulation_type = 'transport', simulation_title = None):
         '''Establish general conditions'''
@@ -89,7 +96,7 @@ class ROSSPkg():
         self.elements = database['elements']
         self.minerals = database['minerals']
 
-    def transport(self, simulation_time, simulation_perspective = None, module_characteristics = {}, cells_per_module = 12, parameterized_timestep = None, kinematic_flow_velocity = None, exchange_factor = 1e10):
+    def transport(self, simulation_time, simulation_perspective = None, module_characteristics = {}, timestep = None, cells_per_module = 12, kinematic_flow_velocity = None, exchange_factor = 1e5):
         '''Define the TRANSPORT block'''
         self.parameters['simulation_time'] = simulation_time
         self.parameters['exchange_factor'] = exchange_factor
@@ -104,7 +111,7 @@ class ROSSPkg():
         self.parameters['module_diameter_mm'] =  201                 
         self.parameters['permeate_tube_diameter_mm'] =  29            
         self.parameters['module_length_m'] =  1.016                 
-        self.parameters['permeate_flow_m3_per_day'] = 40                  
+        self.parameters['permeate_flow_m3_per_hour'] = 40/(day/hour)                  
         self.parameters['max_feed_flow_m3_per_hour'] = 15.9                  
         self.parameters['membrane_thickness_mm'] = 250 * (nano / milli)   
         self.parameters['feed_thickness_mm'] = 0.8636
@@ -146,14 +153,14 @@ class ROSSPkg():
 
         # calculate simulation timestep that adheres to the Courant condition   
         self.parameters['timestep'] = self.parameters['module_length_m'] / feed_velocity  # seconds
-        if parameterized_timestep:
-            self.parameters['timestep'] = parameterized_timestep
+        if timestep:
+            self.parameters['timestep'] = timestep
             
         courant_timestep = self.variables['cell_meters'] / feed_velocity
         if self.parameters['timestep'] < courant_timestep:
             self.parameters['timestep'] = courant_timestep
             
-        self.parameters['permeate_moles_per_cell'] = (self.parameters['permeate_flow_m3_per_day']/day/liter * self.parameters['water_grams_per_liter'] / self.parameters['water_mw']) * (self.parameters['timestep'] / self.parameters['cells_per_module'])      #moles / (cell * self.parameters['timestep'])
+        self.parameters['permeate_moles_per_cell'] = (self.parameters['permeate_flow_m3_per_hour']/hour/liter * self.parameters['water_grams_per_liter'] / self.parameters['water_mw']) * (self.parameters['timestep'] / self.parameters['cells_per_module'])      #moles / (cell * self.parameters['timestep'])
         
         # exit the function for evaporation simulations
         self.results['transport_block'] = []
@@ -173,8 +180,10 @@ class ROSSPkg():
         
         # define the domain-dependent parameters
         domain_line = ''
+        cp_cell_volume_proportion = 0.05 #/self.parameters['cells_per_module']
+        bulk_cell_volume_proportion = 0.95 #/self.parameters['cells_per_module']
         if self.parameters['domain'] == 'dual':
-            domain_line = f'-stagnant\t\t1\t\t{exchange_factor}\t\t\t0.1\t\t0.9 \t # dual domain\n#\t\t\t^stagnant cells\t^exchange factor\t^CP volume\t^bulk volume'
+            domain_line = f'-stagnant\t\t1\t\t{exchange_factor}\t\t\t{cp_cell_volume_proportion}\t\t{bulk_cell_volume_proportion} \t # dual domain\n#\t\t\t^stagnant cells\t^exchange factor\t^CP porosity\t^bulk porosity'
         if self.parameters['domain'] == 'single' or (self.parameters['domain'] == 'dual' and self.parameters['domain_phase'] == 'mobile'):
             first_cell = 1
             final_cell = self.parameters['cells_per_module']*self.parameters['quantity_of_modules']
@@ -207,18 +216,24 @@ class ROSSPkg():
             print('permeate_removal_per_cell', self.parameters['permeate_moles_per_cell'])
             print('active_cm_squared_cell', (self.parameters['active_m2_cell'] / centi**2))
 
-    def reaction(self, permeate_approach = 'linear_permeate', permeate_efficiency = 1, head_loss = 0.89, final_cf = 2, evaporation_steps = 15):     
+    def reaction(self, final_cf = None, permeate_efficiency = 1, head_loss = 0.89, evaporation_steps = 15):     
         '''Define the REACTION blocks'''
         # establish parameters
-        self.parameters['permeate_approach'] = permeate_approach
-        
+        self.parameters['permeate_approach'] = 'linear_permeate'
+        if isnumber(final_cf):
+            self.parameters['permeate_approach'] = 'linear_cf'
+#             module_cf = final_cf**(1/self.parameters['quantity_of_modules'])
+        elif final_cf is not None:
+            print(f'--> ERROR: The final_cf parameter value < {final_cf} > is not a number.')
+    
         cfs = []
         cell_moles = []
         reaction_parameters = []
         self.results['reaction_block'] = []
         self.cumulative_cf = 1 
+        module_previous_moles_removed = 0 
         for module in range(self.parameters['quantity_of_modules']): 
-            if permeate_approach == 'linear_permeate':
+            if self.parameters['permeate_approach'] == 'linear_permeate':
                 initial_moles_removed = self.parameters['permeate_moles_per_cell'] * 2 / (1 + head_loss)
                 final_moles_removed = initial_moles_removed * head_loss
                 removed_moles_slope = ((final_moles_removed - initial_moles_removed) / (self.parameters['cells_per_module'])) / permeate_efficiency
@@ -249,32 +264,57 @@ class ROSSPkg():
                     print('moles_remaining', moles_remaining)
                     print('moles_removed', moles_removed)
 
-            elif permeate_approach == 'linear_cf':
-                module_iteration = module_previous_moles_removed = moles_removed = 0
+            elif self.parameters['permeate_approach'] == 'linear_cf':
+                module_iteration = moles_removed = 0
                 initial_cf = 1
-
-                cf_slope = (final_cf - initial_cf) / self.parameters['cells_per_module']
-                for cell in range(self.parameters['cells_per_module']):
+                cfs = []
+                effective_cells = self.parameters['cells_per_module']*self.parameters['quantity_of_modules']
+                cf_slope = (final_cf - initial_cf) / effective_cells
+                for cell in range(self.parameters['cells_per_module']*(module), self.parameters['cells_per_module']*(module+1)):
                     cell_cf = (cell+1) * cf_slope + initial_cf
-                    cfs.append(cell_cf)    
-
+                    cfs.append(cell_cf)  
                 for cf in cfs:
+                    if self.verbose:
+                        print('\n',)
                     moles_to_be_removed = self.variables['feed_moles'] - (self.variables['feed_moles'] / cf)
                     if module_iteration == 0:
-                        moles = self.variables['feed_moles']
-                        reaction_parameters.append(moles_to_be_removed)
-                    if module_iteration > 0:
+                        if len(reaction_parameters) > 0:
+                            if self.verbose:
+                                print('to be', moles_to_be_removed, '\tlast parameter', reaction_parameters[-1], '\tprevious removal', sum(reaction_parameters))
+                            module_previous_moles_removed += reaction_parameters[-1]
+                        else:
+                            if self.verbose:
+                                print('to be', moles_to_be_removed, )
+                        if module == 0:
+                            reaction_parameters.append(moles_to_be_removed)
+                        else:
+                            reaction_parameters.append(moles_to_be_removed - module_previous_moles_removed)
+                    else:
                         module_previous_moles_removed += reaction_parameters[-1] 
+                        if moles_to_be_removed < module_previous_moles_removed:
+                            print(f'--> ERROR: The reaction parameter is negative: {moles_to_be_removed - module_previous_moles_removed}.')
                         reaction_parameter = moles_to_be_removed - module_previous_moles_removed
+                        if self.verbose:
+                            print('to be', moles_to_be_removed, '\tlast parameter', reaction_parameters[-1], '\tthis parameter', reaction_parameter, '\tprevious removal', module_previous_moles_removed)
                         reaction_parameters.append(reaction_parameter)
-                        moles_removed -= moles_to_be_removed
 
+                    moles_removed = sum(reaction_parameters)
                     module_iteration += 1
 
+                    # verify the CF calculations and efffects in the system
+                    measured_cf = self.variables['feed_moles'] / (self.variables['feed_moles']-moles_removed)
+                    consistency = round(measured_cf, 5) == round(cf, 5)
+                    if not consistency:
+                        print(f'--> ERROR: The measured cf {measured_cf} is dissimilar from the target cf {cf}.') 
+                    else:
+                        if self.verbose:
+                            print('cf consistency between the measured and predicted: ', consistency)
+
                 cf = cfs[-1]
-                self.cumulative_cf *= cf
-                moles_remaining = self.variables['feed_moles'] - moles_removed  
-                self.variables[f'module {module}'] = {'cf': cf}
+                
+            self.cumulative_cf *= cf
+            moles_remaining = self.variables['feed_moles'] - moles_removed  
+            self.variables[f'module {module}'] = {'cf': cf}
 
             if self.parameters['simulation_type'] == 'transport':
                 for cell in range(1, self.parameters['cells_per_module']+1):
@@ -296,15 +336,16 @@ class ROSSPkg():
                 # the calculated reaction parameters will be added and printed to a generated PHREEQC input file
                 final_solution_mass = moles_remaining * self.parameters['water_mw'] * milli  #kg water mass
                 if self.parameters['os'] == 'windows':
-                    self.results['reaction_block'].append('#%s' %(permeate_approach))
-                    if permeate_approach == 'linear permeate':
+                    self.results['reaction_block'].append('# {}'.format(self.parameters['permeate_approach']))
+                    if self.parameters['permeate_approach'] == 'linear_permeate':
                         self.results['reaction_block'].append(f'''
-                #Permeate efficiency parameter: {permeate_efficiency}
-                #Effluent relative pressure: {head_loss}''')
+    #Permeate efficiency parameter: {permeate_efficiency}
+    #Effluent relative pressure: {head_loss}''')
 
-                        self.results['reaction_block'].append(f'''    #Effluent module {module + 1}:
-                #Estimated CF: {sigfigs_conversion(cf, 4)}
-                #Estimated final water mass: {final_solution_mass}\n\n''')
+                    self.results['reaction_block'].append(f'''    
+    #Effluent module {module + 1}:
+#Estimated CF: {sigfigs_conversion(cf, 4)}
+#Estimated final water mass: {final_solution_mass}\n\n''')
 
             elif self.parameters['simulation_type'] == 'evaporation':
                 # define the reaction block
@@ -317,7 +358,7 @@ class ROSSPkg():
             if self.verbose:
                 print(f'Effluent module {module + 1} CF: {cf}')
 
-    def solutions(self, water_selection = '', water_characteristics = {}, solution_description = '', parameterized_alkalinity = False, parameterized_ph_charge = True):
+    def solutions(self, water_selection = '', water_characteristics = {}, solution_description = '', parameterized_ph_charge = True):
         """Specify the SOLUTION block of the simulation."""
         # create the solution line of the input file
         self.results['solution_block'] = []
@@ -338,34 +379,43 @@ class ROSSPkg():
 
         #=============================================================================
         # determine which predefined solution should be simulated
-        
         def define_elements():
             elements_lines = []
             predicted_effluent = {}
             for element, information2 in information.items():
                 self.parameters['solution_elements'].append(element)
+                
                 conc = information2['concentration (ppm)']
                 predicted_effluent[element] = conc*self.cumulative_cf
-                ref = information2['reference']
+                ref = ''
+                if 'reference' in information2:
+                    ref = information2['reference']
+                
+                form = ''
+                if 'form' in information2:
+                    form = 'as {}'.format(information2['form'])
+                    
                 if element in self.elements:                          
                     if len(str(conc)) <= 3:
-                        elements_lines.append(f'{element}\t\t{conc}\t#{ref}')
+                        elements_lines.append(f'{element}\t\t{conc}\t{form}\t#{ref}')
                     else:
-                        elements_lines.append(f'{element}\t{conc}\t#{ref}')
+                        elements_lines.append(f'{element}\t{conc}\t{form}\t#{ref}')
                 else:
                     print('\n--> ERROR: The {} element is not accepted by the {} database'.format(element, self.parameters['database_selection']))
             return elements_lines, predicted_effluent
             
         self.parameters['solution_elements'] = []
         temperature = ph = alkalinity = pe = None
-        if water_selection not in ['', 'custom']:       
+        water_files = glob(os.path.join(self.parameters['root_path'], 'water_bodies','*.json'))
+        water_bodies = [re.search('([a-z\_]+)(?=\.)', file).group() for file in water_files]
+        if water_selection in water_bodies:       
             # import the predefined water body
             water_file_path = os.path.join(self.parameters['root_path'], 'water_bodies', f'{water_selection}.json')
             water_body = json.load(open(water_file_path))
             
             for content, information in water_body.items():
                 if content == 'element':
-                    elements_lines, self.predicted_effluent = define_elements()                                
+                    elements_lines, self.predicted_effluent = define_elements()       
                 elif content == 'temperature':
                     temperature = information['celcius']
                     temperature_reference = information['reference']
@@ -375,27 +425,39 @@ class ROSSPkg():
                 elif content == 'Alkalinity':
                     alkalinity = information['value']
                     alkalinity_reference = information['reference'] 
+                    alkalinity_form = 'as {}'.format(information['form'])    
                 elif content == 'pH':
                     ph = information['value']
                     ph_reference = information['reference']
 
-        if water_characteristics != {}:
+        elif water_characteristics != {}:
             for content, information in water_characteristics.items():
                 if content == 'element':
                     elements_lines, self.predicted_effluent = define_elements()
                 # create the temperature line of the input file
                 elif content == 'temperature':                    
                     temperature = water_characteristics['temperature']['value']
-                    temperature_reference = water_characteristics['temperature']['reference']
+                    temperature_reference = ''
+                    if 'reference' in water_characteristics['temperature']:
+                        temperature_reference = water_characteristics['temperature']['reference']
                 elif content == 'pe':       
                     pe = water_characteristics['pe']['value']
-                    pe_reference = water_characteristics['pe']['reference']
+                    pe_reference = ''
+                    if 'reference' in water_characteristics['pe']:
+                        pe_reference = water_characteristics['pe']['reference']
                 elif content == 'Alkalinity':
                     alkalinity = water_characteristics['Alkalinity']['value']
-                    alkalinity_reference = water_characteristics['Alkalinity']['reference'] 
+                    alkalinity_reference = ''
+                    alkalinity_form = ''
+                    if 'reference' in water_characteristics['Alkalinity']:
+                        alkalinity_reference = water_characteristics['Alkalinity']['reference']
+                    if 'form' in water_characteristics['Alkalinity']:
+                        alkalinity_form = 'as {}'.format(water_characteristics['Alkalinity']['form'])    
                 elif content == 'pH':
                     ph = water_characteristics['pH']['value']
-                    ph_reference = water_characteristics['pH']['reference']
+                    ph_reference = ''
+                    if 'reference' in water_characteristics['pH']:
+                        ph_reference = water_characteristics['pH']['reference']
                     
         # parameterize the lines of the SOLUTIONS block
         temperature_line = ''
@@ -407,14 +469,13 @@ class ROSSPkg():
 
         alkalinity_line = ''
         ph_line = ''
+        if alkalinity:
+            alkalinity_line = f'Alkalinity \t {alkalinity} {alkalinity_form} #{alkalinity_reference}'
         if ph is not None:
-            if parameterized_ph_charge and not parameterized_alkalinity:
+            ph_line = f'pH \t\t {ph} #{ph_reference}'
+            if parameterized_ph_charge and not alkalinity:
                 ph_line = f'pH \t\t {ph} charge #{ph_reference}'
-                alkalinity_line = ''      
-            elif not parameterized_ph_charge:
-                ph_line = f'pH \t\t {ph} #{ph_reference}'
-                if parameterized_alkalinity and alkalinity is not None:
-                    alkalinity_line = f'Alkalinity \t {alkalinity} #{alkalinity_reference}'
+                alkalinity_line = ''
                    
         unit_line = 'units \t ppm' 
         elements_line = '\n'.join(elements_lines)
@@ -508,10 +569,10 @@ class ROSSPkg():
         # create parameter lines             
         if output_filename is None:
             count = 0
-            selected_output_file_name = '_'.join([str(x) for x in [datetime.date.today(), self.parameters['water_selection'], self.parameters['simulation_type'], self.parameters['database_selection'], self.parameters['simulation'], count]]) 
+            selected_output_file_name = '-'.join([str(x) for x in [datetime.date.today(), self.parameters['water_selection'], self.parameters['simulation_type'], self.parameters['database_selection'], self.parameters['simulation'], count]]) 
             while os.path.exists(f'{selected_output_file_name}.txt'):
                 count += 1
-                selected_output_file_name = '_'.join([str(x) for x in [datetime.date.today(), self.parameters['water_selection'], self.parameters['simulation_type'], self.parameters['database_selection'], self.parameters['simulation'], count]]) 
+                selected_output_file_name = '-'.join([str(x) for x in [datetime.date.today(), self.parameters['water_selection'], self.parameters['simulation_type'], self.parameters['database_selection'], self.parameters['simulation'], count]]) 
         else:
             selected_output_file_name = output_filename
 
@@ -584,7 +645,7 @@ class ROSSPkg():
             self.simulation_path = os.path.join(directory, simulation_name)
             os.mkdir(self.simulation_path)
             
-        # define the simulation output path 
+        # define the simulation output path             
         if external_file:
             self.parameters['output_file_name'] = 'output.pqo'
             self.parameters['output_path'] = os.path.join(self.simulation_path, self.parameters['output_file_name'])
@@ -703,7 +764,8 @@ class ROSSPkg():
                 if row != 1:
                     self.parameters['domain_phase'] = 'Mobile'
             elif re.search('H2O -1; ', row):
-                permeate_moles += float(re.sub('(H2O -1; )', '', row))
+                moles = re.search('([0-9]+\.[0-9]+)', row).group()
+                permeate_moles += float(moles)
             elif re.search('TRANSPORT', row):
                 self.parameters['simulation_type'] = 'transport'
             elif re.search('-cells', row):
@@ -746,28 +808,13 @@ class ROSSPkg():
             self.parameters['active_m2'] = 37
         self.parameters['active_m2_cell'] = self.parameters['active_m2']/self.parameters['cells_per_module']
         self.export(simulation_name, input_file_path, external_file = True)
+        
 
     def execute(self, simulated_to_real_time = 9.29):
         '''Execute a PHREEQC input file '''
         def run(input_file, first=False):
             phreeqc = self.phreeqc_mod.IPhreeqc()  
-            print(self.parameters['database_path'])
-            print(type(self.parameters['database_path']))
-            #loading a file
-            error = phreeqc.load_database("/Users/Andrew/Documents/Research/University of Victoria Civil Engineering/ROSSpy/rosspy/databases/unix/pitzer.dat")  
-            print(error)
-            print(phreeqc.get_error_string())
-    
-            # loading a string
-#             db = open(self.parameters['database_path'])
-#             database = '\n'.join([line for line in db])
-#             print(database)
-            phreeqc.load_database_string('solution_master_species')
-            print(dir(phreeqc))
-#             database = '\n'.join([line for line in db])
-            print(phreeqc.get_error_string())
-            
-            # run the simulation
+            phreeqc.load_database(self.parameters['database_path'])
             phreeqc.run_string(input_file)
             
             # define the conc dictionary
@@ -823,7 +870,7 @@ class ROSSPkg():
             
         return self.results['csv_data']
 
-    def process_selected_output(self, selected_output_path = None, plot_title = None, title_font = 'xx-large', label_font = 'x-large', x_label_number = 6, export_name = None, export_format = 'svg', individual_plots = None):
+    def process_selected_output(self, selected_output_path = None, scale_ions = True, plot_title = None, title_font = 'xx-large', label_font = 'x-large', x_label_number = 6, export_name = None, export_format = 'svg', individual_plots = None):
         """Interpreting the PHREEQC SELECTED_OUTPUT file and conducting the plotting functions"""
         if plot_title is not None:
             self.plot_title = plot_title
@@ -885,9 +932,98 @@ class ROSSPkg():
             data = self.brine_plot(pyplot, title_font, label_font, export_format, x_label_number, export_name)
         elif self.parameters['simulation'] == 'scaling':
             data = self.scaling_plot(title_font, label_font, individual_plots, x_label_number, export_name, export_format)
+            if scale_ions:
+                self.ion_proportions(data)
         else:
             print('--> ERROR: The < simulation_perspective > parameter is unpredicted.')
+            
         return data
+    
+    def ion_proportions(self, data):
+        # calculate the masses of each element in each scalant
+        mineral_elements = {}
+        stoich = None
+        total = 0
+        for column in data:
+            mineral = re.search('([A-Za-z]+)', column).group()
+            formula = self.minerals[mineral]['formula']
+            formula = re.sub(':[0-9+]H2O|\(OH\)[0-9]+', '', formula)
+            mineral_elements[mineral] = {}
+
+            index = 0
+            for ch in formula:
+                if re.search('[A-Z]', ch):
+                    element = ch
+                    if re.search('[a-z]', formula[index+1]):
+                        element += formula[index+1]
+                        if len(formula) >= index+2+1:
+                            if re.search('[0-9]', formula[index+2]):
+                                stoich = float(formula[index+2])
+                            else:
+                                stoich = 1
+                    elif re.search('[0-9]', formula[index+1]):
+                        stoich = float(formula[index+1])
+                    else:
+                        stoich = 1
+                elif re.search('[0-9]', ch):
+                    continue
+
+                if element in elemental_masses:
+                    if stoich:
+                        mass = elemental_masses[element] * stoich
+                        total += mass
+                        mineral_elements[mineral][element] = mass
+                        stoich = None
+                        element = None
+
+                index += 1
+
+            mineral_elements[mineral]['total_grams'] = total
+
+        # determine the ionic proportions of the respective mineral, without solvated water or hydroxide
+        elemental_ratio = {}
+        for mineral in mineral_elements:
+            elemental_ratio[mineral] = {}
+            for element in mineral_elements[mineral]:
+                elemental_ratio[mineral][element] = mineral_elements[mineral][element] / mineral_elements[mineral]['total_grams']
+
+        # calculate the precipitated mass of each element
+        scale_ratio = {}
+        elements = {}
+        for mineral in elemental_ratio:
+            index = 0
+            scale_ratio[mineral] = {}
+            unit = '(g/m^2)'
+            if self.parameters['simulation_type'] == 'evaporation':
+                unit = '(g)'
+            for scale in data[f'{mineral} {unit}']:  
+                df_index = f'{data.index[index]} (m)'
+                scale_ratio[mineral][df_index] = {}
+                if df_index not in elements:
+                    elements[df_index] = {}
+                    elements[df_index]['ion (g/m^2)'] = {}
+                for element in elemental_ratio[mineral]:
+                    scale_ratio[mineral][df_index][element] = elemental_ratio[mineral][element] * float(scale)
+
+                    if element in elements[df_index]['ion (g/m^2)']:
+                        elements[df_index]['ion (g/m^2)'][element] += scale_ratio[mineral][df_index][element]
+                    elif element in elements[df_index]:
+                        elements[df_index][element] += scale_ratio[mineral][df_index][element]
+                    else:
+                        if element == 'total_grams':
+                            elements[df_index][element] = scale_ratio[mineral][df_index][element]
+                        else:
+                            elements[df_index]['ion (g/m^2)'][element] = scale_ratio[mineral][df_index][element]
+
+                index += 1
+
+        if self.verbose:
+            pprint(elements)
+
+        # export the parsed scale ions
+        ionic_path = os.path.join(self.simulation_path, 'scale_ions.json')
+        with open(ionic_path, 'w') as output:
+            json.dump(elements, output, indent = 4)
                                  
     def brine_plot(self, pyplot, title_font, label_font, export_format, x_label_number, export_name):
         """Generate plots of the elemental concentrations from effluent brine in the PHREEQC SELECTED_OUTPUT file"""
@@ -900,6 +1036,7 @@ class ROSSPkg():
         # parse the brine concentrations from the raw data
         pyplot.figure(figsize = (17,10))
         non_zero_elements = []
+        non_zero_columns = []
         time = initial_solution_time = 0
         data = {} 
         concentration_serie = []
@@ -923,19 +1060,25 @@ class ROSSPkg():
         
         for element in columns:  
             stripped_element = re.search('([A-Z][a-z]?(?:\(\d\))?(?=\(mol\/kgw\)))', element).group()
-            non_zero_elements.append(stripped_element)
+            if self.results['csv_data'][element].iloc[-1] > 0: 
+                non_zero_elements.append(stripped_element)
+                non_zero_columns.append(element)
             concentration_serie = []
             if self.parameters['simulation_perspective'] == 'all_time':
                 time_serie = []
                 data[element] = {}
+                insufficient_elements = set()
                 for index, row in self.results['csv_data'].iterrows():
-                    if all(row[element] > 1e-16 for element in columns):
+                    if all(row[element] > 1e-16 for element in non_zero_columns):
                         concentration_serie.append(row[element])
                         time = sigfigs_conversion(row['time'], 3)
                         time_serie.append(time) # - initial_solution_time * self.parameters['timestep'])
                         data[element][time] = sigfigs_conversion(row[element], 4)
                     else:
                         initial_solution_time += 1
+                        for element in non_zero_columns:
+                            if row[element] < 1e-16:
+                                insufficient_elements.add(element)
                 pyplot.plot(time_serie,concentration_serie)
                                     
             elif self.parameters['simulation_perspective'] == 'all_distance':
@@ -959,8 +1102,12 @@ class ROSSPkg():
                             data[f'{element} (mol)'][sigfig_x] = sigfigs_conversion(concentration_serie[index], 4)
                     else:
                         concentration_serie.append(row[element])
-                        distance_serie.append(row['dist_x'])                       
-                
+                        distance_serie.append(row['dist_x'])       
+                        
+        if self.parameters['simulation_perspective'] == 'all_time':
+            if len(time_serie) == 0:
+                    print(f'\n\n--> ERROR: The {insufficient_elements} elements remain below the 1E-16 Molal threshold.')
+                        
         # define the brine plot
         if self.parameters['simulation_perspective'] == 'all_time':
             x_location = []
@@ -1148,7 +1295,7 @@ class ROSSPkg():
         non_zero_minerals = set()
         self.variables['precipitated_minerals'] = {}
         for column in self.results['csv_data'].columns:
-            if re.search('([A-Z][a-z]{2,})', column) and not re.search('[_(]', column):
+            if re.search('([A-Z][a-z]{2,})', column) and not re.search('[_(]|(?:Metal)', column):
                 mineral = re.search('([A-Z][a-z]{2,})', column).group()
                 csv_minerals.append(mineral)
                 for value in self.results['csv_data'][mineral]:
